/****************************************************************************
**	File:	TLICHAT.C
**
**	Description:  	This is a simple DOS program that illustrates the use
**                of TLI calls with a conversation between two workstations.
**						The first workstation must start the program in the
**						server mode.  The second workstation will then start
**						the program in the client mode.  Each time a string is
**						entered, followed by a carriage return, the string is
**						sent to the other workstation and displayed on its
**						screen.
**
**
**		DISCLAIMER
**
**	Novell, Inc. makes no representations or warranties with respect to
**	any NetWare software, and specifically disclaims any express or
**	implied warranties of merchantability, title, or fitness for a
**	particular purpose.
**
**	Distribution of any NetWare software is forbidden without the
**	express written consent of Novell, Inc.  Further, Novell reserves
**	the right to discontinue distribution of any NetWare software.
**
**	Novell is not responsible for lost profits or revenue, loss of use
**	of the software, loss of data, costs of re-creating lost data, the
**	cost of any substitute equipment or program, or claims by any party
**	other than you.  Novell strongly recommends a backup be made before
**	any software is installed.   Technical support for this software
**	may be provided at the discretion of Novell.
**
**	Programmers:
**
**		Ini	Who						Firm
**		-----------------------------------------------------------------------
**		MDP	Marina Pimentel				Novell Developer Support.
**
**	History:
**
**		When		Who	What
**		-----------------------------------------------------------------------
**		2-2-96	        MDP	First code.
*/

/****************************************************************************
**	Include headers, macros, function prototypes, etc.
*/

/*------------------------------------------------------------------------
**	MACROS
*/

#define NWDOS

#define SPX_SOCKET 31  	 /* arbitrary socket */

#ifndef t_errno
	int t_errno;
#endif


/*------------------------------------------------------------------------
**	ANSI
*/


#include <time.h>
#include <io.h>
#include <dos.h>
#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <process.h>
#include <string.h>
#include <fcntl.h>


/*------------------------------------------------------------------------
**	NetWare
*/


#include <sap.h>
#include <tispxipx.h>
#include <tiuser.h>
#include <nwcalls.h>


/*------------------------------------------------------------------------
**	Prototypes
*/

int   SPXDisconReason( int fd );
int   listen(int fd, struct t_call tcall,int nfd);
int   init(int fd, struct t_call tcall,int* nfd);
void	StartAdvertise(void);
void	StartQuery(void);
int 	StartTLI(int* fd);
int 	ConnectTLI(int fd);

/*------------------------------------------------------------------------
**	Globals
*/

char  		nfdOpen = 0;
char  		fdOpen = 0;
char  		initDone = 0;
char  		isServer = 0;
WORD			sType = 0x9001;
IPX_ADDR		spx_addr;
struct		t_bind tbind;
SPX_OPTS		spx_options;
struct		t_call tcall;

/****************************************************************************
** Function: Main (void)
*/

void main ( int argc, char *argv[] )
{
	int			ccode = 0;
	int	 		fd = -1;
	int			nfd = -1;
	ECB			*rECB;
	IPXHeader   *rHeader;
	char        c,*cPtr;

	if(strcmp(strupr(argv[1]),"S") && strcmp(strupr(argv[1]),"C"))
	{
	  printf ("tlichat s or tlichat c\n");
	  exit(1);
	}

	if(!strcmp(strupr(argv[1]),"S"))
	{
		isServer=1;
		StartAdvertise();
	}/* end ifServer */
	else
		StartQuery();

	ccode = StartTLI(&fd);
	if(ccode)
		goto cleanup;

	if(!isServer)
	{
		ccode = ConnectTLI(fd);
		if(ccode)
			goto cleanup;
	}/* end if client */

	ccode = listen(fd,tcall,nfd);

	cleanup:
	if(isServer)
	{
		printf("Shutting down SAP\n");
		ccode=ShutdownSAP();
		printf("ShutdownSAP returned ccode [%X]\n",ccode);
	}
	if(fdOpen)
		t_close(fd);
	if(nfdOpen)
		t_close(nfd);
}
/* End of main  */

/****************************************************************************
** Function: SPXDisconReason (int fd)
** This function determines the reason that the SPX connection terminated.
*/

int SPXDisconReason(int fd)
{
	int		ccode =0;
	struct	t_discon discon;
	char		msg[50]="\0";

	ccode = t_rcvdis(fd, &discon);
	if (ccode == -1)
	{
		t_error( "t_rcvdis failed" );
		return -1;
	}
	switch( discon.reason )
	{
		case TLI_SPX_CONNECTION_FAILED:
			strcpy(msg,"Connection failed");
			break;
		case TLI_SPX_CONNECTION_TERMINATED:
			strcpy(msg,"Connection terminated by client");
			break;
		case TLI_SPX_MALFORMED_PACKET:
			strcpy(msg,"Internal SPX interface error -- malformed packet");
			break;
		default:
			strcpy(msg,"Unknown termination reason");
	 }
	 printf( "SPX Connection terminated: %s\n", msg );
	 return 0;
}
/* end of SPXDisconReason() */

/****************************************************************************
** Function: listen(int fd, struct t_call tcall,int nfd)
** If isServer, t_listen is called until something is heard, or ~ is pressed.
** If client or connection is setup with client, messages can be sent and
** accepted.
*/

int listen(int fd, struct t_call tcall,int nfd)
{
	int		ccode =0;
	int 		listenResult = -1,flags,len;
	char 		buf_1[100];
	char		buf2[ 100 ]="\0";
	char 		message [ 100 ] = "\0";
	int 		bytesSent = -1;

	printf("Press ~ to exit program\n");
	do
	{
		if((isServer) && (!initDone))
		do
		{
			if(kbhit())
			 {
				gets(message);
				strncpy(buf2,message,98);
				strcat(buf2,"\0");
				if(!strcmp(buf2,"~"))
					return 0;
				else
					printf("A second workstation must run \"tlichat c\" before you can chat.\n");
			 }

		} while ( ((listenResult = t_listen(fd, &tcall)) == -1) && (t_errno == TNODATA));

		if ( initDone || (!isServer))
		{
			strcpy(buf2,"\0");
			if(kbhit())
			{
				if ( strcmp(buf2,"~") )
				{
					gets(message);
					strncpy(buf2,message,98);
					strcat(buf2,"\0");
					if(strcmp(buf2,"~"))
					{
						do
						{
							bytesSent = t_snd(nfd,buf2,strlen(buf2)+1,0);
						} while((bytesSent == -1)&&(t_errno == TFLOW));
						if(bytesSent == -1)
							printf("Some problem occurred and did not send\n");
					}/* end else !~ */
				}/* end else !~ */
			}/* end if kbhit() */
		}/* end if !isServer and initDone */

		if(isServer && !initDone)
		{
			ccode = init(fd,tcall,&nfd);
			if(ccode == -1)
			{
				printf("failed to initialize connection \n");
				return -1;
			}
			initDone = 1;
		}/* initDone */
		if(!isServer && !initDone)
		{
			nfd = fd;
			initDone = 1;
		}

/* initialization of connection is completed, now both sides could receive */

		flags = 0;		/* initialize for t_rcv call */
		strcpy(buf_1,"\0");
		while(t_rcv(nfd,buf_1,sizeof(buf_1),&flags)!=-1)
			printf("msg:  %s\n",buf_1);

	} while (strcmp(buf2,"~"));

	if ((t_errno == TLOOK) && (t_look( fd ) == T_DISCONNECT))
	{
		ccode = SPXDisconReason(fd);
		if (ccode == -1)
	 return -1;
	}
	else
	{
		if(!(t_errno == TNODATA))
			t_error( "t_errno returned:  " );
		return -1;
	}
	return 0;
}

/****************************************************************************
** Function: init(int fd, struct t_call tcall,int* nfd)
** This function is called if isServer and the connection with the
** client is not yet established.  This call establishes the connection.
*/

int init(int fd, struct t_call tcall,int* nfd)
{
	int		ccode =0;

	*nfd = t_open("/dev/nspx", O_RDWR | O_NDELAY, (struct t_info *)0);
	if (*nfd == -1)
	{
		t_error( "t_open after listen failed" );
		return -1;
	}
	nfdOpen=1;
	ccode = t_bind(*nfd, (struct t_bind *)0, (struct t_bind *)0);
	if(ccode == -1)
	{
		t_error( "t_bind failed" );
		return -1;
	}
	ccode = t_accept(fd, *nfd, &tcall);
	if (ccode == -1)
	{
		t_error( "t_accept failed\n" );
		return -1;
	}
	printf( "Connection opened with client -- ready to chat\n" );
	return 0;
}


/****************************************************************************
** Function: StartAdvertise (void)
** This function starts the TLIserver sapping.
*/

void StartAdvertise(void)
{
	char		sName[48];
	BYTE		sSocket[2];
	int		ccode = 0;

	strcpy(sName,"TLIServer");
	*(WORD*)&sSocket[0]=0x9000;
	ccode=AdvertiseService(sType,sName,sSocket);
	if(ccode==0xFF || ccode==0xF0)
	{
		printf("Exiting...AdvertiseService returned [%X]\n",ccode);
		exit(1);
	}
	if (ccode != 0)
		printf("Completion code was [%X]\n",ccode);
}
/* end of StartAdvertise() */

/****************************************************************************
** Function: StartQuery(void)
** This call makes the query to find any TLIServer that is sapping.
*/

void StartQuery(void)
{
	int		ccode = 0;
	WORD 		qType, rSize;
	SAP		sBuf;

	qType = 0x3;
	rSize = 96;

	ccode = QueryServices (qType, sType, rSize, &sBuf);
	if (ccode)
	{
		printf("Could not find a workstation running \"tlichat s\"\n");
		printf("Please start \"tlichat s\" on another workstation on LAN");
		exit(1);
	}

	ccode = IPXInitialize();

	memcpy(spx_addr.ipxa_net, sBuf.Network, 4);
	memcpy(spx_addr.ipxa_node, sBuf.Node,6);

}
/* end of StartQuery() */

/****************************************************************************
** Function: StartTLI(int* fd)
** This call does preliminary setup for the TLI connection.
*/

int StartTLI(int* fd)
{
	int			ccode = 0;

	*fd = t_open( "/dev/nspx", O_RDWR | O_NDELAY, ( struct t_info * )0 );
	if (*fd == -1)
	{
		t_error( "open of /dev/nspx failed" );
		return -1;
	}
	fdOpen=1;

	if(!isServer)
	{

/* No need to bind to a specific socket */

		ccode = t_bind( *fd, ( struct t_bind * )0, ( struct t_bind * )0);
		if(ccode == -1)
		{
			t_error( "bind failed" );
			return -1;
		}
	}

/* initialize spx_addr and tbind structures */

	spx_addr.ipxa_socket[ 0 ]  = 0;
	spx_addr.ipxa_socket[ 1 ]  = SPX_SOCKET;
	if(isServer)
	{
		tbind.addr.len 	      = sizeof( spx_addr );
		tbind.addr.maxlen	      = sizeof( spx_addr );
		tbind.addr.buf	      	= ( char * ) &spx_addr;
		tbind.qlen		      	= 5;

		ccode = t_bind( *fd, &tbind, &tbind );
		if (ccode == -1)
		{
			t_error( "bind failed" );
			return -1;
		}
	}
	tcall.addr.buf 		         			= ( char * ) &spx_addr;
	tcall.addr.maxlen  	               	= sizeof( spx_addr );
	tcall.addr.len	 	         				= sizeof( spx_addr );
	spx_options.spx_connectionID[ 0 ]		= 0;
	spx_options.spx_connectionID[ 1 ]		= 0;
	spx_options.spx_allocationNumber[ 0 ] 	= 0;
	spx_options.spx_allocationNumber[ 1 ] 	= 0;
	tcall.opt.buf		         				= ( char * )&spx_options;
	tcall.opt.len	  	         				= sizeof( spx_options );
	tcall.opt.maxlen   	                 	= sizeof( spx_options );
	tcall.udata.buf		         			= ( char * )0;
	tcall.udata.maxlen 	                 	= 0;
	tcall.udata.len		         			= 0;

	return 0;
}
/* end of StartTLI() */

/****************************************************************************
** Function: ConnectTLI(int fd)
** This function is called when the client side needs to establish
** the connection with the server.
*/

int ConnectTLI(int fd)
{
	int			rval = -1;
	int			ccode = 0;

	ccode = t_connect( fd, &tcall, &tcall );
	if(ccode == -1)
	{
		do
		{
			rval = t_rcvconnect(fd, &tcall);
		}while(rval == -1 && t_errno == TNODATA);
	}/* end if t_connect returned -1 */
	if(!(((t_errno == TLOOK) && (t_look(fd) == T_CONNECT)) || rval == 0 ))
	{
		rval = t_errno;
		t_error("t_rcvconnect failed");
		return -1;
	}
	return 0;
}
/* end of ConnectTLI() */